import pytest

from passpie import clipboard


def test_clipboard_on_osx_ensure_commands(mocker):
    mocker.patch('passpie.clipboard.process.call')
    mock_ensure_commands = mocker.patch('passpie.clipboard.ensure_commands')
    commands = clipboard.OSX_COMMANDS

    clipboard._copy_osx('text')

    mock_ensure_commands.assert_called_once_with(commands)


def test_clipboard_on__ensure_commands(mocker):
    mocker.patch('passpie.clipboard.process.call')
    mock_ensure_commands = mocker.patch('passpie.clipboard.ensure_commands')
    commands = clipboard.LINUX_COMMANDS

    clipboard._copy_linux('text')

    mock_ensure_commands.assert_called_once_with(commands)


def test_copy_calls_copy_osx_when_on_darwin_system(mocker):
    mocker.patch('passpie.clipboard.process.call')
    mocker.patch('passpie.clipboard.platform.system', return_value='Darwin')
    mock_copy_osx = mocker.patch('passpie.clipboard._copy_osx')
    mock_copy_linux = mocker.patch('passpie.clipboard._copy_linux')
    mock_copy_windows = mocker.patch('passpie.clipboard._copy_windows')

    clipboard.copy('text')

    assert mock_copy_osx.called is True
    assert mock_copy_linux.called is False
    assert mock_copy_windows.called is False
    mock_copy_osx.assert_called_once_with('text', 0)


def test_copy_calls_copy_linux_when_on_linux_system(mocker):
    mocker.patch('passpie.clipboard.process.call')
    mocker.patch('passpie.clipboard.platform.system', return_value='Linux')
    mock_copy_osx = mocker.patch('passpie.clipboard._copy_osx')
    mock_copy_linux = mocker.patch('passpie.clipboard._copy_linux')
    mock_copy_windows = mocker.patch('passpie.clipboard._copy_windows')

    clipboard.copy('text')

    assert mock_copy_linux.called is True
    assert mock_copy_osx.called is False
    assert mock_copy_windows.called is False
    mock_copy_linux.assert_called_once_with('text', 0)


def test_copy_calls_copy_windows_when_on_windows_system(mocker):
    mocker.patch('passpie.clipboard.process.call')
    mocker.patch('passpie.clipboard.platform.system', return_value='Windows')
    mock_copy_osx = mocker.patch('passpie.clipboard._copy_osx')
    mock_copy_linux = mocker.patch('passpie.clipboard._copy_linux')
    mock_copy_windows = mocker.patch('passpie.clipboard._copy_windows')

    clipboard.copy('text')

    assert mock_copy_windows.called is True
    assert mock_copy_osx.called is False
    assert mock_copy_linux.called is False
    mock_copy_windows.assert_called_once_with('text', 0)


def test_copy_calls_copy_cygwin_when_on_cygwin_system(mocker):
    mocker.patch('passpie.clipboard.platform.system', return_value='cygwin system')
    mock_copy_cygwin = mocker.patch('passpie.clipboard._copy_cygwin')
    text = 's3cr3t'

    clipboard.copy(text)

    assert mock_copy_cygwin.called
    mock_copy_cygwin.assert_called_once_with(text, 0)


def test_logs_error_msg_when_platform_not_supported(mocker):
    mocker.patch('passpie.clipboard.platform.system', return_value='unknown')
    mock_logger = mocker.patch('passpie.clipboard.logging')

    clipboard.copy('text')
    assert mock_logger.error.called
    msg = "platform 'unknown' copy to clipboard not supported"
    mock_logger.error.assert_called_once_with(msg)


def test_ensure_commands_logs_error_when_command_not_found(mocker):
    mocker.patch('passpie.clipboard.which', return_value=False)
    mock_logging = mocker.patch('passpie.clipboard.logging')
    clipboard.ensure_commands(clipboard.LINUX_COMMANDS)

    assert mock_logging.error.called


def test_ensure_commands_returns_command(mocker):
    commands = {'xclip': ['xclip']}
    mocker.patch('passpie.clipboard.which', return_value=True)

    result = clipboard.ensure_commands(commands)

    assert result == commands['xclip']


def test_clear_sleep_for_delay_seconds(mocker):
    mocker.patch('passpie.clipboard.process')
    mocker.patch('passpie.clipboard.sys.stdout')
    mocker.patch('passpie.clipboard.print', create=True)
    mock_time = mocker.patch('passpie.clipboard.time')

    clipboard.clean('some command', delay=5)
    assert mock_time.sleep.call_count == 5


def test_clear_calls_command_to_clear_with_whitespace_char(mocker):
    mocker.patch('passpie.clipboard.sys.stdout')
    mocker.patch('passpie.clipboard.print', create=True)
    mocker.patch('passpie.clipboard.time')
    mock_process = mocker.patch('passpie.clipboard.process')

    clipboard.clean('some command', delay=5)
    assert mock_process.call.called
    mock_process.call.assert_called_once_with('some command', input='\b')


def test_clear_is_called_when_clear_is_passed_to_copy_osx(mocker):
    mocker.patch('passpie.clipboard.ensure_commands', return_value='command')
    mocker.patch('passpie.clipboard.process')
    mock_clean = mocker.patch('passpie.clipboard.clean')

    clipboard._copy_osx('text', clear=5)
    assert mock_clean.called
    mock_clean.assert_called_once_with('command', delay=5)


def test_clear_is_called_when_clear_is_passed_to_copy_linux(mocker):
    mocker.patch('passpie.clipboard.ensure_commands', return_value='command')
    mocker.patch('passpie.clipboard.process')
    mock_clean = mocker.patch('passpie.clipboard.clean')

    clipboard._copy_linux('text', clear=5)
    assert mock_clean.called
    mock_clean.assert_called_once_with('command', delay=5)



import pytest
from passpie.process import Proc, call, DEVNULL, PIPE


@pytest.fixture
def mock_popen(mocker):
    return mocker.patch('passpie.process.Popen')


def test_call_uses_proc_communicate_with_input(mocker, mock_popen):
    MockProc = mocker.patch('passpie.process.Proc')
    MockProc().__enter__.return_value.communicate.return_value = ('', '')
    call(['echo', 'hello'])
    assert MockProc().__enter__.called is True
    assert MockProc().__exit__.called is True


def test_call_sets_stderr_as_devnull_when_logger_level_is_not_debug(mocker, mock_popen):
    MockProc = mocker.patch('passpie.process.Proc')
    MockProc().__enter__.return_value.communicate.return_value = ('', '')
    mock_logging = mocker.patch('passpie.process.logging')
    mock_logging.getLogger().getEffectiveLevel.return_value = 0
    mock_logging.DEBUG = 10

    call(['echo', 'hello'])
    args, kwargs = MockProc.call_args
    assert kwargs['stderr'] == DEVNULL


def test_call_sets_stderr_as_pipe_when_logger_level_is_set_to_debug(mocker, mock_popen):
    MockProc = mocker.patch('passpie.process.Proc')
    MockProc().__enter__.return_value.communicate.return_value = ('', '')
    mock_logging = mocker.patch('passpie.process.logging')
    mock_logging.getLogger().getEffectiveLevel.return_value = 10
    mock_logging.DEBUG = 10

    call(['echo', 'hello'])
    args, kwargs = MockProc.call_args
    assert kwargs['stderr'] == PIPE


def test_call_output_and_error_are_utf8_decoded(mocker, mock_popen):
    MockProc = mocker.patch('passpie.process.Proc')
    output = mocker.MagicMock()
    error = mocker.MagicMock()
    MockProc().__enter__.return_value.communicate.return_value = (output, error)

    result_output, result_error = call(['echo', 'hello'])

    assert output.decode.called is True
    assert error.decode.called is True
    assert result_output == output.decode('utf-8')
    assert result_error == error.decode('utf-8')


def test_call_output_and_error_doesnt_decode_in_case_of_unicode(mocker, mock_popen):
    MockProc = mocker.patch('passpie.process.Proc')
    output = mocker.MagicMock()
    error = mocker.MagicMock()
    output.decode.side_effect = AttributeError
    error.decode.side_effect = AttributeError
    MockProc().__enter__.return_value.communicate.return_value = (output, error)

    result_output, result_error = call(['echo', 'hello'])

    assert result_output == output
    assert result_error == error



from datetime import datetime
import os

import click

from passpie.importers import BaseImporter
from passpie.credential import make_fullname

try:
    from pysswords.db import Database

    def found_pysswords():
        return True
except ImportError:
    def found_pysswords():
        return False


class PysswordsImporter(BaseImporter):

    def match(self, filepath):
        if not found_pysswords():
            self.log('Pysswords is not installed')
            return False

        try:
            assert os.path.isdir(filepath)
            assert '.keys' in os.listdir(filepath)
        except AssertionError:
            self.log('.keys not found in path')
            return False

        return True

    def handle(self, filepath):
        db = Database(path=filepath)
        passphrase = click.prompt('Pysswords passphrase', hide_input=True)
        if not db.check(passphrase):
            return []

        credentials = []
        for cred in db.credentials:
            credential_dict = {
                'fullname': make_fullname(cred.login, cred.name),
                'name': cred.name,
                'login': cred.login,
                'password': db.decrypt(cred.password, passphrase),
                'comment': cred.comment,
                'modified': datetime.now(),
            }
            credentials.append(credential_dict)

        return credentials



import passpie.config
import yaml


def mock_open():
    try:
        from mock import mock_open as mopen
    except:
        from unittest.mock import mock_open as mopen
    return mopen()


def test_config_read_opens_path_and_load_yaml_content(mocker):
    config_file = mocker.patch('passpie.config.open', mock_open(), create=True)()
    mock_yaml = mocker.patch('passpie.config.yaml')

    passpie.config.read('path')
    assert mock_yaml.load.called
    mock_yaml.load.assert_called_once_with(config_file.read())


def test_config_read_logs_debug_when_config_file_not_found_and_returns_default(mocker):
    mocker.patch('passpie.config.open', side_effect=IOError, create=True)
    mock_logging = mocker.patch('passpie.config.logging')

    result = passpie.config.read('path')
    assert result == passpie.config.DEFAULT
    assert mock_logging.debug.called
    mock_logging.debug.assert_called_once_with('config file "path" not found')


def test_config_read_logs_error_when_config_file_malformed_found_and_returns_empty(mocker):
    mocker.patch('passpie.config.open', side_effect=yaml.scanner.ScannerError, create=True)
    mock_logging = mocker.patch('passpie.config.logging')

    result = passpie.config.read('path')
    assert result == {}
    assert mock_logging.error.called
    mock_logging.error.assert_called_once_with('Malformed user configuration file: path')


def test_config_load_gets_settings_from_local_config_when_exists(mocker):
    expected_config = {
        'short_commands': True
    }
    mocker.patch('passpie.config.os.path.exists', return_value=True)
    mocker.patch('passpie.config.read', return_value=expected_config)

    configuration = passpie.config.load()
    assert configuration is not None
    assert configuration['short_commands'] is expected_config['short_commands']

def test_config_load_overrides_global_config_with_local_config_vars(mocker):
    global_config = {
        'genpass_pattern': "[\w]{16}",
        'short_commands': True
    }
    local_config = {
        'genpass_pattern': "[\w]{32}"
    }
    mocker.patch('passpie.config.os.path.exists', return_value=True)
    mocker.patch('passpie.config.read', return_value=local_config)
    mocker.patch('passpie.config.read_global_config', return_value=global_config)

    configuration = passpie.config.load()
    assert configuration is not None
    assert configuration['genpass_pattern'] == local_config['genpass_pattern']
    assert configuration['short_commands'] == global_config['short_commands']

def test_config_load_has_default_config_loaded(mocker):
    default_config = passpie.config.DEFAULT
    global_config = {
        'genpass_pattern': "[\w]{16}",
        'short_commands': True
    }
    local_config = {
        'genpass_pattern': "[\w]{32}"
    }
    mocker.patch('passpie.config.os.path.exists', return_value=True)
    mocker.patch('passpie.config.read', return_value=local_config)
    mocker.patch('passpie.config.read_global_config', return_value=global_config)

    configuration = passpie.config.load()
    assert configuration is not None
    assert configuration['path'] == default_config['path']
    assert configuration['key_length'] == default_config['key_length']
    assert configuration['table_format'] == default_config['table_format']

def test_config_load_has_default_config_loaded(mocker):
    default_config = passpie.config.DEFAULT
    global_config = {
        'genpass_pattern': "[\w]{16}",
        'short_commands': True
    }
    local_config = {
        'genpass_pattern': "[\w]{32}"
    }
    mocker.patch('passpie.config.os.path.exists', return_value=True)
    mocker.patch('passpie.config.read', return_value=local_config)
    mocker.patch('passpie.config.read_global_config', return_value=global_config)

    configuration = passpie.config.load()
    assert configuration is not None
    assert configuration['path'] == default_config['path']
    assert configuration['key_length'] == default_config['key_length']
    assert configuration['table_format'] == default_config['table_format']


def test_config_load_with_overrides_override_loaded_config(mocker):
    global_config = {
        'genpass_pattern': "[\w]{16}",
        'short_commands': True
    }
    local_config = {
        'genpass_pattern': "[\w]{32}"
    }

    short_commands=False
    genpass_pattern=10
    mocker.patch('passpie.config.os.path.exists', return_value=True)
    mocker.patch('passpie.config.read', return_value=local_config)
    mocker.patch('passpie.config.read_global_config', return_value=global_config)

    overrides = dict(short_commands=short_commands, genpass_pattern=genpass_pattern)
    configuration = passpie.config.load(**overrides)
    assert configuration is not None
    assert configuration['genpass_pattern'] == genpass_pattern
    assert configuration['short_commands'] == short_commands


def test_config_create_adds_an_empty_dot_config_file_to_path_when_default_false(mocker):
    config_file = mocker.patch('passpie.config.open', mock_open(), create=True)()
    mock_yaml_dump = mocker.patch('passpie.config.yaml.dump')
    overrides = {}
    passpie.config.create('path', overrides)

    config_file.write.assert_called_once_with(mock_yaml_dump(overrides, default_flow_style=False))



BASH = """
function _passpie()
{
    COMPREPLY=()
    local word="${COMP_WORDS[COMP_CWORD]}"

    if [ "$COMP_CWORD" -eq 1 ]; then
        COMPREPLY=( $(compgen -W "{commands}" -- "$word") )
    else
        local words=("${COMP_WORDS[@]}")
        unset words[0]
        unset words[$COMP_CWORD]
        COMPREPLY=( $(compgen -W "$(grep -EhriIo '[A-Z0-9._%+-]+@[A-Z0-9.-]+(@[A-Z0-9_\-\.]+)?' {config_path})${IFS}\
                                  $(ls -1 {config_path})" -- "$word") )
    fi
}
complete -F _passpie 'passpie'
"""

FISH = """
function _passpie_credentials
  grep -EhriIo '[A-Z0-9._%+-]+@[A-Z0-9.-]+(@[A-Z0-9_\-\.]+)?' {config_path}
end

function _passpie_credential_names
  ls -1 {config_path}
end

function _passpie_needs_command
  set cmd (commandline -opc)
  if [ (count $cmd) -eq 1 -a $cmd[1] = 'passpie' ]
    return 0
  end
  return 1
end

function _passpie_using_command
  set cmd (commandline -opc)
  if [ (count $cmd) -gt 1 ]
    for arg in $argv
      if [ $arg = $cmd[2] ]
        return 0
      end
    end
  end
  return 1
end

complete -f -c passpie -n '_passpie_needs_command' -a '{commands}' --description 'Manage a credential'
complete -f -c passpie -n '_passpie_using_command {commands}' -a '(_passpie_credentials)
                                                                  (_passpie_credential_names)' --description 'Credential'
"""

ZSH = """
if [[ ! -o interactive ]]; then
    return
fi

compctl -K _passpie passpie

_passpie() {
  local words completions
  read -cA words

  if [ "${#words}" -eq 2 ]; then
    completions="{commands}"
  else
    completions="$(grep -EhriIo '\\b[A-Z0-9._%+-]+@[A-Z0-9.-]+(@[A-Z0-9_\-\.]+)?\\b' {config_path})${IFS}\
                 $(ls -1 {config_path})"
  fi

  reply=(${(ps:\n:)completions})
}
"""

SHELLS = ['zsh', 'fish', 'bash']


def script(shell_name, config_path, commands):
    text = ''
    if shell_name == 'zsh':
        text = ZSH.replace('{commands}', '\n'.join(commands))
        text = text.replace('{config_path}', config_path)
    elif shell_name == 'fish':
        text = FISH.replace('{commands}', ' '.join(commands))
        text = text.replace('{config_path}', config_path)
    elif shell_name == 'bash':
        text = BASH.replace('{commands}', ' '.join(commands))
        text = text.replace('{config_path}', config_path)

    return text



import pytest

from passpie.utils import genpass, mkdir_open, ensure_dependencies, touch


def mock_open():
    try:
        from mock import mock_open as mopen
    except:
        from unittest.mock import mock_open as mopen
    return mopen()


def test_genpass_generates_a_password_with_length_32(mocker):
    password = genpass("\w{32}")
    assert len(password) == 32


def test_mkdir_open_makedirs_on_path_dirname(mocker):
    mock_os = mocker.patch("passpie.utils.os")
    mocker.patch("passpie.utils.open", mock_open(), create=True)
    path = "path/to/foo.pass"
    with mkdir_open(path, "w"):
        dirname = mock_os.path.dirname(path)
        mock_os.makedirs.assert_called_once_with(dirname)


def test_mkdir_open_handle_oserror_for_file_exist(mocker):
    mock_os = mocker.patch("passpie.utils.os")
    mocker.patch("passpie.utils.open", mock_open(), create=True)
    path = "path/to/foo.pass"

    mock_os.makedirs.side_effect = OSError(17, "File Exists")
    with mkdir_open(path, "w") as fd:
        assert fd is not None

    mock_os.makedirs.side_effect = OSError(2, "File Not Found")
    with pytest.raises(OSError):
        with mkdir_open(path, "w") as fd:
            pass


def test_ensure_dependencies_raises_runtime_when_gpg_not_installed(mocker):
    mocker.patch('passpie.utils.which', return_value=None)

    with pytest.raises(RuntimeError) as excinfo:
        ensure_dependencies()

    assert 'GnuPG not installed. https://www.gnupg.org/' == str(excinfo.value)


def test_touch_open_file_on_path(mocker):
    mock_builtin_open = mocker.patch("passpie.utils.open",
                                     mock_open(),
                                     create=True)
    path = 'path'
    touch(path)

    assert mock_builtin_open.called
    mock_builtin_open.assert_called_once_with(path, 'w')



import tempfile

from passpie import completion


def test_script_returns_zsh_script_when_zsh_shell_name(mocker):
    shell_name = 'zsh'
    config_path = tempfile.gettempdir()
    commands = ['add', 'remove', 'update', 'remove', 'search']
    text = completion.script(shell_name=shell_name,
                             config_path=config_path,
                             commands=commands)

    for line in completion.ZSH.split('\n')[:3]:
        assert line in text


def test_script_returns_fish_script_when_fish_shell_name(mocker):
    shell_name = 'fish'
    config_path = tempfile.gettempdir()
    commands = ['add', 'remove', 'update', 'remove', 'search']
    text = completion.script(shell_name=shell_name,
                             config_path=config_path,
                             commands=commands)

    for line in completion.FISH.split('\n')[9:28]:
        assert line in text


def test_script_returns_bash_script_when_bash_shell_name(mocker):
    shell_name = 'bash'
    config_path = tempfile.gettempdir()
    commands = ['add', 'remove', 'update', 'remove', 'search']
    text = completion.script(shell_name=shell_name,
                             config_path=config_path,
                             commands=commands)

    for line in completion.BASH.split('\n')[:3]:
        assert line in text



import yaml
from yaml.reader import ReaderError
from yaml.scanner import ScannerError

from passpie.importers import BaseImporter


class DefaultImporter(BaseImporter):

    def match(self, filepath):
        try:
            with open(filepath) as fp:
                file_content = fp.read()
        except OSError:
            return False

        try:
            dict_content = yaml.load(file_content)
        except (ReaderError, ScannerError):
            return False

        try:
            assert dict_content.get('handler') == 'passpie'
            assert isinstance(dict_content.get('version'), float)
        except AssertionError:
            return False

        return True

    def handle(self, filepath):
        with open(filepath) as fp:
            file_content = fp.read()
        dict_content = yaml.load(file_content)
        credentials = dict_content.get('credentials')
        return credentials



from collections import namedtuple
import os
import shutil
import sys
import tempfile

import yaml

from passpie.importers import find_importer, BaseImporter, get_instances
from passpie.importers.default_importer import DefaultImporter
from passpie.importers.pysswords_importer import PysswordsImporter


def mock_open():
    try:
        from mock import mock_open as mopen
    except:
        from unittest.mock import mock_open as mopen
    return mopen()


def test_find_importer_returns_first_match_default_importer(mocker):
    mock_importer = mocker.Mock()
    mock_importer2 = mocker.Mock()
    mock_importer.match.return_value = False
    mock_importer2.match.return_value = True

    mocker.patch('passpie.importers.get_instances',
                 return_value=[mock_importer, mock_importer2])

    importer = find_importer('mockpath')

    assert importer is mock_importer2


def test_default_importer_match_passpie_exported_yaml(mocker):
    dict_content = {'handler': 'passpie', 'version': 1.0}
    mocker.patch('passpie.importers.default_importer.open',
                 mock_open(), create=True)
    mocker.patch('passpie.importers.default_importer.yaml.load',
                 return_value=dict_content)

    result = DefaultImporter().match('filepath')
    assert result is True


def test_default_importer_returns_false_when_bad_file(mocker):
    mockopen = mocker.patch(
        'passpie.importers.default_importer.open', mock_open(), create=True)()
    mockopen.read.side_effect = OSError

    result = DefaultImporter().match('filepath')
    assert result is False


def test_default_importer_returns_false_when_missing_version_key(mocker):
    dict_content = {'handler': 'passpie'}
    mocker.patch('passpie.importers.default_importer.open',
                 mock_open(), create=True)
    mocker.patch('passpie.importers.default_importer.yaml.load',
                 return_value=dict_content)

    result = DefaultImporter().match('filepath')
    assert result is False


def test_default_importer_returns_false_when_missing_handler_key(mocker):
    dict_content = {'version': 1.0}
    mocker.patch('passpie.importers.default_importer.open',
                 mock_open(), create=True)
    mocker.patch('passpie.importers.default_importer.yaml.load',
                 return_value=dict_content)

    result = DefaultImporter().match('filepath')
    assert result is False


def test_default_importer_returns_false_when_version_keys_isnt_float(mocker):
    dict_content = {'version': '1.0'}
    mocker.patch('passpie.importers.default_importer.open',
                 mock_open(), create=True)
    mocker.patch('passpie.importers.default_importer.yaml.load',
                 return_value=dict_content)

    result = DefaultImporter().match('filepath')
    assert result is False


def test_default_importer_returns_loaded_credentials_from_yaml_file(mocker):
    dict_content = {'credentials': {'name': 'foo', 'name': 'bar'}}
    mocker.patch('passpie.importers.default_importer.open',
                 mock_open(), create=True)
    mocker.patch('passpie.importers.default_importer.yaml.load',
                 return_value=dict_content)

    result = DefaultImporter().handle('filepath')
    assert result is dict_content.get('credentials')


def test_default_importer_match_returns_false_when_bad_yaml(mocker):
    mocker.patch('passpie.importers.default_importer.open',
                 mock_open(), create=True)
    mocker.patch('passpie.importers.default_importer.yaml.load',
                 side_effect=[yaml.scanner.ScannerError])

    result = DefaultImporter().match('filepath')
    assert result is False


def test_base_importer_log_calls_logging_debug_with_message(mocker):
    mock_logging = mocker.patch('passpie.importers.logging')
    importer = BaseImporter()
    message = 'test debugging'
    importer.log(message)

    assert mock_logging.debug.called is True
    mock_logging.debug.assert_called_once_with(message)


def test_get_instances_returns_instances_of_all_found_importers(mocker):
    Importer = mocker.Mock()
    Importer2 = mocker.Mock()
    Importer3 = mocker.Mock()
    mocker.patch('passpie.importers.get_all',
                 return_value=[Importer, Importer2, Importer3])

    importers = list(get_instances())
    assert len(importers) == 3
    assert Importer() in importers
    assert Importer2() in importers
    assert Importer3() in importers


def test_pysswords_returns_false_with_logging_when_not_installed(mocker):
    to_patch = 'passpie.importers.pysswords_importer.found_pysswords'
    mocker.patch(to_patch, return_value=False)
    importer = PysswordsImporter()
    importer.log = mocker.Mock()

    result = importer.match('filepath')
    assert result is False
    assert importer.log.called
    importer.log.assert_called_once_with('Pysswords is not installed')


def test_pysswords_returns_false_with_logging_when_path_not_dir(mocker):
    to_patch = 'passpie.importers.pysswords_importer.found_pysswords'
    mocker.patch(to_patch, return_value=True)
    mock_os = mocker.patch('passpie.importers.pysswords_importer.os')
    mock_os.path.is_dir.return_value = False
    importer = PysswordsImporter()
    importer.log = mocker.Mock()

    result = importer.match('filepath')
    assert result is False
    assert importer.log.called
    importer.log.assert_called_once_with('.keys not found in path')


def test_pysswords_returns_false_with_logging_when_keys_not_in_path(mocker):
    to_patch = 'passpie.importers.pysswords_importer.found_pysswords'
    mocker.patch(to_patch, return_value=True)
    mock_os = mocker.patch('passpie.importers.pysswords_importer.os')
    mock_os.path.is_dir.return_value = True
    mock_os.listdir.return_value = []
    importer = PysswordsImporter()
    importer.log = mocker.Mock()

    result = importer.match('filepath')
    assert result is False
    assert importer.log.called
    importer.log.assert_called_once_with('.keys not found in path')


def test_pysswords_match_returns_true_when_expected_path(mocker):
    to_patch = 'passpie.importers.pysswords_importer.found_pysswords'
    mocker.patch(to_patch, return_value=True)
    mock_os = mocker.patch('passpie.importers.pysswords_importer.os')
    mock_os.path.is_dir.return_value = True
    mock_os.listdir.return_value = ['.keys']
    importer = PysswordsImporter()

    result = importer.match('filepath')
    assert result is True


def test_pysswords_handle_returns_empty_when_bad_passphrase(mocker):
    mocker.patch('passpie.importers.pysswords_importer.click')
    to_patch = 'passpie.importers.pysswords_importer.Database'
    mock_pysswords_db = mocker.patch(to_patch, create=True)()
    mock_pysswords_db.check.return_value = False
    importer = PysswordsImporter()

    result = importer.handle('path')
    assert result == []


def test_pysswords_handle_returns_pysswords_credentials(mocker):
    Cred = namedtuple('Credential', 'name login password comment')
    credentials = [
        Cred('name', 'login', 'password', 'comment')
    ]
    mocker.patch('passpie.importers.pysswords_importer.click')
    mocker.patch('passpie.importers.pysswords_importer.make_fullname')
    to_patch = 'passpie.importers.pysswords_importer.Database'
    mock_pysswords_db = mocker.patch(to_patch, create=True)()
    mock_pysswords_db.check.return_value = True
    mock_pysswords_db.credentials = credentials
    importer = PysswordsImporter()

    result = importer.handle('path')
    assert credentials[0].name in [c['name'] for c in result]


def test_find_importers_through_entry_points(mocker):
    from passpie import importers

    temp_dir = tempfile.mkdtemp()
    sys.path.insert(0, temp_dir)

    with open(os.path.join(temp_dir, 'fake_module.py'), 'w') as f:
        f.write("""\
from passpie.importers import BaseImporter

class FakeKeepassImporterClass(BaseImporter):
    pass
""")

    import pkg_resources
    fake_ep = pkg_resources.EntryPoint(
        'fake_keepass', 'fake_module',
        attrs=('FakeKeepassImporterClass', ))
    mock_iter_entry_points = mocker.patch(
        'pkg_resources.iter_entry_points',
        return_value=iter([fake_ep, ]))

    try:
        target_klass = None
        for klass in importers._get_importers_from_entry_points():
            if klass.__name__ == 'FakeKeepassImporterClass':
                target_klass = klass
                break

        mock_iter_entry_points.assert_called_once_with('passpie_importers')

        assert target_klass.__name__ == 'FakeKeepassImporterClass'
        assert target_klass.__module__ == 'fake_module'

    finally:
        sys.path.remove(temp_dir)
        shutil.rmtree(temp_dir, ignore_errors=True)


def test_get_all_yields_importers_from_entry_points(mocker):
    from passpie import importers

    class FakeImporter(importers.BaseImporter):
        pass
    fake_importers = {FakeImporter, }

    mocker.patch.object(importers, '__all__', new=[])
    mock_ep_finder = mocker.patch.object(
        importers, '_get_importers_from_entry_points',
        return_value=iter(fake_importers))

    found_importers = set(importers.get_all())
    assert found_importers == fake_importers



import pytest

from passpie.credential import split_fullname, make_fullname


def test_split_fullname_raises_value_error_when_invalid_name(mocker):
    invalid_fullname = ""
    with pytest.raises(ValueError):
        split_fullname(invalid_fullname)


def test_split_fullname_returns_expected_login_and_name(mocker):
    assert split_fullname("foo@example") == ("foo", "example")
    assert split_fullname("foo@example.com") == ("foo", "example.com")
    assert split_fullname("@example.com") == ("", "example.com")
    assert split_fullname("example.com") == (None, "example.com")
    assert split_fullname("foo@example.com@archive.com") == ("foo@example.com", "archive.com")


def test_make_fullname_returns_expected_fullname(mocker):
    assert make_fullname("foo", "bar") == "foo@bar"
    assert make_fullname("_", "bar") == "_@bar"
    assert make_fullname(None, "bar") == "@bar"



import inspect
import os
import glob
import logging

import pkg_resources


class BaseImporter(object):

    def log(self, message):
        logging.debug(message)

    def match(self, filepath):
        """Return True if file for filepath parseable by importer
        """
        raise NotImplementedError('Handle should be implemented by base class')

    def handle(self, filepath):
        """Return all credentials decrypted from filepath
        """
        raise NotImplementedError('Handle should be implemented by base class')


__all__ = [os.path.basename(f)[:-3]
           for f in glob.glob(os.path.dirname(__file__) + "/*.py")]


def _import_all_importer_files():
    __import__(__name__, globals(), locals(), __all__, 0)


def _get_importers_from_entry_points():
    for ep in pkg_resources.iter_entry_points('passpie_importers'):
        try:
            module = __import__(ep.module_name, globals(), locals(), ep.attrs, 0)
            klass = getattr(module, ep.attrs[0])
            if klass is not BaseImporter and issubclass(klass, BaseImporter):
                yield klass
        except (ImportError, AttributeError):
            pass


def get_all():
    """Get all subclasses of BaseImporter from module and return and generator
    """

    _import_all_importer_files()

    for module in (value for key, value in globals().items()
                   if key in __all__):
        for klass_name, klass in inspect.getmembers(module, inspect.isclass):
            if klass is not BaseImporter and issubclass(klass, BaseImporter):
                yield klass

    for klass in _get_importers_from_entry_points():
        yield klass


def get_instances():
    """Get instances of all found detected importer classes"""
    return (wclass() for wclass in get_all())


def find_importer(filepath):
    for importer in get_instances():
        if importer.match(filepath):
            return importer



import re


def split_fullname(fullname):
    regex = re.compile(r'(?:(?P<login>.+?(?:\@.+?)?)@(?P<name>.+?$))')
    regex_name_only = re.compile(r'(?P<at>@)?(?P<name>[\w\-\_\.]+$)')

    if regex.match(fullname):
        mobj = regex.match(fullname)
    elif regex_name_only.match(fullname):
        mobj = regex_name_only.match(fullname)
    else:
        raise ValueError("Not a valid name")

    if mobj.groupdict().get('at'):
        login = ""
    else:
        login = mobj.groupdict().get('login')
    name = mobj.groupdict().get("name")

    return login, name


def make_fullname(login, name):
    fullname = "{}@{}".format("" if login is None else login, name)
    return fullname



from passpie.table import Table


def test_table_colorile_calls_click_style_with_key_on_text(mocker):
    mock_click = mocker.patch('passpie.table.click')
    table = Table(headers=['name', 'login'], colors={'red': 'red'})
    table.colorize(key='red', text='text')
    assert mock_click.style.called
    mock_click.style.assert_called_once_with('text', fg=table.colors.get('red'))


def test_render_hide_with_starts_hidden_columns(mocker):
    table = Table(headers=['password'], hidden=['password'])
    mocker.patch('passpie.table.click.style', return_value='password')
    mock_tabulate = mocker.patch('passpie.table.tabulate')
    data = [{'password': 's3cr3t'}, {'password': 'another, s3cr3t'}]

    table.render(data)
    assert mock_tabulate.called
    mock_tabulate.assert_called_once_with([['*****'], ['*****']],
                                          ['password'],
                                          tablefmt=table.table_format,
                                          missingval=table.missing,
                                          numalign='left')



def test_render_colorize_expected_columns(mocker):
    colors = {'name': 'red', 'login': 'blue'}
    table = Table(headers=['name', 'login'], colors=colors)
    mock_tabulate = mocker.patch('passpie.table.tabulate')
    mocker.patch.object(table, 'colorize', return_value='colorized')
    mocker.patch('passpie.table.click.style', return_value='header')
    data = [{'login': 'foo', 'name': 'example.com'}]
    rows = [
        ['colorized', 'colorized']
    ]

    table.render(data)
    assert mock_tabulate.called
    assert table.colorize.called
    table.colorize.assert_any_call('name', 'example.com')
    table.colorize.assert_any_call('login', 'foo')
    mock_tabulate.assert_called_once_with(rows,
                                          ['header', 'header'],
                                          tablefmt=table.table_format,
                                          missingval=table.missing,
                                          numalign='left')



from tinydb import where, Query
from tinydb.storages import MemoryStorage

from passpie.database import Database, is_repo_url, PasspieStorage
from .helpers import MockerTestCase


class StorageTests(MockerTestCase):

    def setUp(self):
        self.mock_os = self.patch('passpie.database.os')
        self.mock_shutil = self.patch("passpie.database.shutil")
        self.storage = PasspieStorage("path")

    def test_read_returns_all_found_credentials_in_default_dict(self):
        mock_open = self.patch("passpie.database.open",
                               self.mock_open(), create=True)
        mock_open().read.return_value = "{}"
        self.mock_os = self.patch('passpie.database.os')
        self.mock_os.walk.return_value = [
            ('/foo', ('bar',), ('baz',)),
            ('/foo/bar', (), ('eggs.pass',)),
            ('/foo/bar2', (), ('spam.pass',))
        ]
        storage = PasspieStorage("path")
        storage.write = self.Mock()
        elements = storage.read()
        self.assertIn("_default", elements.keys())
        self.assertEqual(len(elements["_default"]), 2)

    def test_write_writes_all_elements_from_data_default_dict(self):
        mock_mkdir_open = self.patch("passpie.database.mkdir_open",
                                     self.mock_open(), create=True)

        self.patch("passpie.database.PasspieStorage.read",
                   return_value={"_default": {}})
        data = {"_default": {1: {"name": "example", "login": "foo"},
                             2: {"name": "example", "login": "bar"}}}
        storage = PasspieStorage("path")

        storage.write(data)

        self.assertEqual(mock_mkdir_open.call_count, 2)

    def test_delete_removes_credentials_files(self):
        credentials = [
            {"name": "bar", "login": "foo"},
            {"name": "spam", "login": "foozy"},
        ]

        self.storage.delete(credentials)

        for cred in credentials:
            credpath = self.mock_os.path.join(cred["name"], cred["login"])
            self.mock_os.remove.assert_any_call(credpath)

    def test_delete_removes_empty_directories(self):
        credentials = [
            {"name": "bar", "login": "foo"},
            {"name": "spam", "login": "foozy"},
        ]
        self.mock_os.listdir.side_effect = [[], ["not empty"]]
        self.storage.delete(credentials)

        credpath = self.mock_os.path.join(credentials[0]["name"],
                                          credentials[0]["login"])
        self.mock_shutil.rmtree.assert_called_once_with(
            self.mock_os.path.dirname(credpath))


def test_database_initialization_creates_tempdir_and_clone_when_path_is_repo_url(mocker):
    assert is_repo_url('https://foo@example.com/user/repo.git')
    assert is_repo_url('https://github.com/marcwebbie/passpie.git')
    assert is_repo_url('git@github.com:marcwebbie/passpie.git')
    assert not is_repo_url('http://example.com')
    assert not is_repo_url(None)
    assert not is_repo_url('')
    assert not is_repo_url('++++++++++++++')


def test_database_has_keys_returns_true_when_file_dot_keys_found_in_db_path(mocker):
    config = {
        'path': 'path',
        'extension': '.pass',
    }
    mock_exists = mocker.patch('passpie.database.os.path.exists', return_value=True)
    mock_join = mocker.patch('passpie.database.os.path.join')
    db = Database(config)

    assert db.has_keys() is True
    mock_exists.assert_called_once_with(mock_join('path', '.keys'))


def test_database_credential_with_fullname_does_a_db_where_with_split_fullname(mocker):
    config = {
        'path': 'path',
        'extension': '.pass',
    }
    db = Database(config)
    mocker.patch('passpie.database.split_fullname', return_value=('login', 'name'))
    mock_get = mocker.patch.object(db, 'get', return_value=[{}])

    result = db.credential('login@name')
    assert db.get.called
    assert result == [{}]
    db.get.assert_called_once_with((where('login') == 'login') & (where('name') == 'name'))


def test_database_add_insert_credential_to_database(mocker):
    config = {
        'path': 'path',
        'extension': '.pass',
    }
    db = Database(config)
    mock_get = mocker.patch.object(db, 'insert')
    mocker.patch('passpie.database.split_fullname', return_value=('login', 'name'))
    mock_datetime = mocker.patch('passpie.database.datetime')
    credential = dict(fullname='login@name',
                      name='name',
                      login='login',
                      password='password',
                      comment='comment',
                      modified=mock_datetime.now())

    db.add(fullname='login@name', password='password', comment='comment')
    assert db.insert.called
    db.insert.assert_called_once_with(credential)


def test_database_add_with_empty_login_logs_error_and_return_none(mocker):
    config = {
        'path': 'path',
        'extension': '.pass',
    }
    db = Database(config)
    mocker.patch('passpie.database.split_fullname', return_value=(None, 'name'))
    mock_logging = mocker.patch('passpie.database.logging')

    result = db.add(fullname='login@name', password='password', comment='comment')
    assert result is None
    mock_logging.error.assert_called_once_with(
        'Cannot add credential with empty login. use "@<name>" syntax')


def test_database_update_uses_table_update_credential_to_database(mocker):
    config = {
        'path': 'path',
        'extension': '.pass',
    }
    db = Database(config)
    mocker.patch.object(db, 'table', mocker.MagicMock())
    mocker.patch('passpie.database.make_fullname', return_value='login@name')
    mock_datetime = mocker.patch('passpie.database.datetime')
    values = {
        'login': 'login',
        'name': 'name',
        'comment': 'new comment'
    }

    db.update(fullname='login@name', values=values)

    assert db.table().update.called
    db.table().update.assert_called_once_with(values, (where('fullname') == 'login@name'))


def test_database_remove_uses_table_remove_credential_from_database(mocker):
    config = {
        'path': 'path',
        'extension': '.pass',
    }
    db = Database(config)
    mocker.patch.object(db, 'table', mocker.MagicMock())
    mocker.patch('passpie.database.make_fullname', return_value='login@name')

    db.remove(fullname='login@name')

    assert db.table().remove.called
    db.table().remove.assert_called_once_with((where('fullname') == 'login@name'))


def test_credentials_returns_sorted_list_credentials(mocker):
    config = {
        'path': 'path',
        'extension': '.pass',
    }
    db = Database(config)
    mocker.patch.object(db, 'all', mocker.MagicMock())
    mock_sorted = mocker.patch('passpie.database.sorted', create=True)

    credentials = db.credentials()
    assert credentials == mock_sorted()


def test_credentials_filter_credentials_by_login_and_name_when_full_fullname_passed(mocker):
    config = {
        'path': 'path',
        'extension': '.pass',
    }
    db = Database(config)
    mocker.patch('passpie.database.split_fullname', return_value=('foo', 'example.com'))
    mocker.patch.object(db, 'search')
    mocker.patch.object(db, 'all')
    Credential = Query()

    credentials = db.credentials(fullname="foo@example.com")
    assert db.search.called is True
    assert db.all.called is False
    db.search.assert_called_once_with(
        (Credential.login == 'foo') & (Credential.name == 'example.com')
    )


def test_credentials_filter_credentials_by_login_and_name_when_empty_login_fullname_passed(mocker):
    config = {
        'path': 'path',
        'extension': '.pass',
    }
    db = Database(config)
    mocker.patch.object(db, 'search')
    mocker.patch.object(db, 'all')
    mocker.patch('passpie.database.split_fullname', return_value=('', 'example.com'))
    Credential = Query()

    credentials = db.credentials(fullname="@example.com")
    assert db.search.called is True
    assert db.all.called is False
    db.search.assert_called_once_with(
        (Credential.login == "") & (Credential.name == "example.com")
    )


def test_credentials_filter_credentials_by_login_and_name_when_name_only_fullname_passed(mocker):
    config = {
        'path': 'path',
        'extension': '.pass',
    }
    db = Database(config)
    mocker.patch.object(db, 'search')
    mocker.patch.object(db, 'all')
    mocker.patch('passpie.database.split_fullname', return_value=(None, 'example.com'))
    Credential = Query()

    credentials = db.credentials(fullname="@example.com")
    assert db.search.called is True
    assert db.all.called is False
    db.search.assert_called_once_with(
        (Credential.name == "example.com")
    )


def test_database_matches_uses_table_remove_credential_from_database(mocker):
    config = {
        'path': 'path',
        'extension': '.pass',
    }
    db = Database(config)
    mocker.patch.object(db, 'search', mocker.MagicMock())
    mocker.patch('passpie.database.make_fullname', return_value='login@name')
    regex = '.*'
    Credential = Query()
    result = db.matches(regex=regex)

    assert db.search.called
    assert isinstance(result, list)
    db.search.assert_called_once_with(
        Credential.name.matches(regex) |
        Credential.login.matches(regex) |
        Credential.comment.matches(regex)
    )



import pytest
from passpie.history import ensure_git, Repository, clone


@pytest.fixture
def mock_process(mocker):
    return mocker.patch('passpie.history.process')


def test_ensure_git_logs_debug_for_git_not_installed(mocker, mock_process):
    mocker.patch('passpie.history.which', return_value=None)
    mock_logging = mocker.patch('passpie.history.logging')

    @ensure_git()
    def func():
        pass
    result = func()

    assert mock_logging.debug.called
    assert result is None


def test_ensure_returns_specified_return_value_when_exception(mocker, mock_process):
    mocker.patch('passpie.history.which', return_value='/usr/bin/gpg')
    mock_logging = mocker.patch('passpie.history.logging')
    return_value = 'return value'

    @ensure_git(return_value=return_value)
    def func():
        raise ValueError()
    result = func()

    assert result == return_value
    assert mock_logging.debug.called


def test_git_init_creates_a_repository_on_path(mocker, mock_process):
    path = 'some_path'
    cmd = ['git', 'init', path]
    repo = Repository(path)
    repo.init()

    mock_process.call.assert_called_once_with(cmd, cwd=repo.path)


def test_calls_pull_rebase_on_initialization_when_autopull_is_passed(mocker, mock_process):
    mocker.patch.object(Repository, 'pull_rebase')
    autopull = ['origin', 'master']
    repo = Repository('path', autopull)
    assert repo.pull_rebase.called
    repo.pull_rebase.assert_called_once_with(*autopull)


def test_git_pull_rebase_calls_expected_command(mocker, mock_process):
    cmd = ['git', 'pull', '--rebase', 'origin', 'master']
    repo = Repository('path')
    repo.pull_rebase()

    mock_process.call.assert_called_once_with(cmd, cwd=repo.path)


def test_git_pull_rebase_calls_expected_command_when_remote_branch_is_passed(mocker, mock_process):
    cmd = ['git', 'pull', '--rebase', 'another_origin', 'another_branch']
    repo = Repository('path')
    repo.pull_rebase(remote='another_origin', branch='another_branch')

    mock_process.call.assert_called_once_with(cmd, cwd=repo.path)


def test_git_clone_calls_expected_command_with_repository_url(mocker, mock_process):
    mock_tempdir = mocker.patch('passpie.history.tempdir', return_value='tempdir')
    dest = 'tempdir'
    url = 'https://foo@example.com/user/repo.git'
    cmd = ['git', 'clone', url, dest]
    clone(url)

    assert mock_tempdir.called
    mock_process.call.assert_called_once_with(cmd)


def test_git_clone_calls_expected_command_with_repository_url_and_destination(mocker, mock_process):
    mock_tempdir = mocker.patch('passpie.history.tempdir', return_value='tempdir')
    url = 'https://foo@example.com/user/repo.git'
    dest = "some/path"
    cmd = ['git', 'clone', url, dest]
    clone(url, dest)

    assert not mock_tempdir.called
    mock_process.call.assert_called_once_with(cmd)


def test_git_push_calls_expected_command(mocker, mock_process):
    cmd = ['git', 'push', 'origin', 'master']
    repo = Repository('path')
    repo.push()

    mock_process.call.assert_called_once_with(cmd, cwd=repo.path)


def test_git_push_calls_expected_command_when_remote_branch_is_passed(mocker, mock_process):
    cmd = ['git', 'push', 'another_origin', 'another_branch']
    repo = Repository('path')
    repo.push(remote='another_origin', branch='another_branch')

    mock_process.call.assert_called_once_with(cmd, cwd=repo.path)


def test_git_add_has_call_with_expected_command(mocker, mock_process):
    cmd = ['git', 'add', '.']
    repo = Repository('path')
    repo.add()

    mock_process.call.assert_called_once_with(cmd, cwd=repo.path)


def test_git_add_has_call_with_expected_command_with_all_flag_when_all_is_true(mocker, mock_process):
    cmd = ['git', 'add', '--all', '.']
    repo = Repository('path')
    repo.add(all=True)

    mock_process.call.assert_called_once_with(cmd, cwd=repo.path)


def test_git_sha_list_has_call_with_expected_command(mocker, mock_process):
    output = 'a\nb'
    mock_process.call.return_value = (output, '')
    cmd = ['git', 'log', '--reverse', '--pretty=format:%h']
    repo = Repository('path')
    result = repo.sha_list()

    assert result == output.splitlines()
    mock_process.call.assert_called_once_with(cmd, cwd=repo.path)


def test_git_commit_creates_commit_with_message(mocker, mock_process):
    message = 'Initial commit'
    cmd = ['git', 'commit', '-m', message]
    repo = Repository('path')
    mocker.patch.object(repo, 'add')

    repo.commit(message)

    repo.add.assert_called_once_with(all=True)
    mock_process.call.assert_called_once_with(cmd, cwd=repo.path)


def test_git_commit_calls_push_when_autopush_set(mocker, mock_process):
    message = 'Initial commit'
    cmd = ['git', 'commit', '-m', message]
    repo = Repository('path', autopush=['origin', 'master'])
    mocker.patch.object(repo, 'push')

    repo.commit(message)
    assert repo.push.called


def test_git_commit_list_has_expected_commit_list(mocker, mock_process):
    commit_list = [
        'another commit',
        'initial commit'
    ]
    mock_process.call.return_value = ("\n".join(commit_list), 'no error')
    cmd = ['git', 'log', '--reverse', '--pretty=format:%s']
    repo = Repository('path')
    commits = repo.commit_list()

    assert len(repo.commit_list()) == len(commits)
    mock_process.call.assert_any_call(cmd, cwd=repo.path)


def test_reset_doesnt_call_git_reset_hard_on_commit_when_not_found(mocker, mock_process):
    sha_list = []
    index = 0

    repo = Repository('path')
    mocker.patch.object(repo, 'sha_list', return_value=sha_list)
    repo.reset(index)

    mock_process.call.called is False


def test_reset_call_git_reset_hard_on_commit_when_found(mocker, mock_process):
    sha_list = [
        'd6b52b5',
        '34c5439',
        '5910c2b',
    ]
    index = 0
    cmd = ['git', 'reset', '--hard', sha_list[index]]

    repo = Repository('path')
    mocker.patch.object(repo, 'sha_list', return_value=sha_list)
    repo.reset(index)

    mock_process.call.assert_called_once_with(cmd, cwd=repo.path)



import json
import logging
import os
import shutil

import click
from click.decorators import Group
import yaml

from . import clipboard, completion, config, checkers, importers
from .crypt import create_keys, encrypt, decrypt
from .database import Database, is_repo_url
from .table import Table
from .utils import genpass, ensure_dependencies
from .history import clone


__version__ = "1.1.0"
pass_db = click.make_pass_decorator(Database)


class AliasedGroup(click.Group):

    def get_command(self, ctx, cmd_name):
        rv = click.Group.get_command(self, ctx, cmd_name)
        if rv is not None:
            return rv
        matches = [x for x in self.list_commands(ctx)
                   if x.startswith(cmd_name)]
        if not matches:
            return None
        elif len(matches) == 1:
            return click.Group.get_command(self, ctx, matches[0])
        ctx.fail('Too many matches: %s' % ', '.join(sorted(matches)))


def ensure_passphrase(passphrase, config):
    encrypted = encrypt('OK', recipient=config['recipient'], homedir=config['homedir'])
    decrypted = decrypt(encrypted,
                        recipient=config['recipient'],
                        passphrase=passphrase,
                        homedir=config['homedir'])
    if not decrypted == 'OK':
        message = "Wrong passphrase"
        message_full = "Wrong passphrase for recipient: {} in homedir: {}".format(
            config['recipient'],
            config['homedir'],
        )
        logging.error(message_full)
        raise click.ClickException(click.style(message, fg='red'))


def validate_remote(ctx, param, value):
    if value:
        try:
            remote, branch = value.split('/')
            return (remote, branch)
        except ValueError:
            raise click.BadParameter('remote need to be in format <remote>/<branch>')


@click.group(invoke_without_command=True,
             cls=AliasedGroup if config.load()['short_commands'] else Group)
@click.option('-D', '--database', help='Database path or url to remote repository',
              envvar="PASSPIE_DATABASE")
@click.option('--autopull', help='Autopull changes from remote pository',
              callback=validate_remote, envvar="PASSPIE_AUTOPULL")
@click.option('--autopush', help='Autopush changes to remote pository',
              callback=validate_remote, envvar="PASSPIE_AUTOPUSH")
@click.option('-v', '--verbose', help='Activate verbose output', count=True)
@click.version_option(version=__version__)
@click.pass_context
def cli(ctx, database, autopull, autopush, verbose):
    try:
        ensure_dependencies()
    except RuntimeError as e:
        raise click.ClickException(click.style(str(e), fg='red'))

    # Override configuration
    config_overrides = {}
    if database:
        config_overrides['path'] = clone(url=database) if is_repo_url(database) else database
    if autopush:
        config_overrides['autopush'] = autopush
    if autopull:
        config_overrides['autopull'] = autopull

    # Setup configuration
    configuration = config.load(**config_overrides)

    # Setup database
    db = Database(configuration)
    ctx.obj = db

    # Verbose
    if verbose == 1:
        logging_level = logging.INFO
    elif verbose > 1:
        logging_level = logging.DEBUG
    else:
        logging_level = logging.CRITICAL
    logging.basicConfig(format="%(levelname)s:passpie.%(module)s:%(message)s",
                        level=logging_level)

    # List credentials
    if ctx.invoked_subcommand is None:
        credentials = db.credentials()
        if credentials:
            table = Table(
                configuration['headers'],
                table_format=configuration['table_format'],
                colors=configuration['colors'],
                hidden=['password']
            )
            click.echo(table.render(credentials))


@cli.command(help='Generate completion scripts for shells')
@click.argument('shell_name', type=click.Choice(completion.SHELLS),
                default=None, required=False)
@click.option('--commands', default=None)
def complete(shell_name, commands):
    commands = ['add', 'copy', 'remove', 'search', 'update']
    script = completion.script(shell_name, config.path, commands)
    click.echo(script)


@cli.command(help="Initialize new passpie database")
@click.option('--force', is_flag=True, help="Force overwrite database")
@click.option('--no-git', is_flag=True, help="Don't create a git repository")
@click.option('--recipient', help="Keyring default recipient")
@click.option('--passphrase', help="Database passphrase")
@pass_db
def init(db, force, no_git, recipient, passphrase):
    if force:
        if os.path.isdir(db.path):
            shutil.rmtree(db.path)
            logging.info('removed directory %s' % db.path)
        elif os.path.isfile(db.path):
            os.remove(db.path)
            logging.info('removed file %s' % db.path)

    try:
        os.makedirs(db.path)
    except (SystemError, OSError):
        message = "Path exists '{}'. `--force` to overwrite".format(db.path)
        raise click.ClickException(click.style(message, fg='yellow'))

    if recipient:
        logging.info('create .passpierc file at %s' % db.path)
        config.create(db.path, defaults=dict(recipient=recipient))
    else:
        logging.info('create .passpierc file at %s' % db.path)
        config.create(db.path, defaults={})
        if not passphrase:
            passphrase = click.prompt('Passphrase',
                                      hide_input=True,
                                      confirmation_prompt=True)
        keys_filepath = os.path.join(db.config['path'], '.keys')
        create_keys(passphrase, keys_filepath, key_length=db.config['key_length'])

    if not no_git:
        logging.info('init git repository in %s' % db.path)
        db.repo.init()
        db.repo.commit(message='Initialized git repository', add=True)

    click.echo("Initialized database in {}".format(db.path))


@cli.command(help='Add new credential to database')
@click.argument("fullname")
@click.option('-p', '--password', help="Credential password")
@click.option('-r', '--random', is_flag=True, help="Randonly generate password")
@click.option('-P', '--pattern', help="Random password regex pattern")
@click.option('-c', '--comment', default="", help="Credential comment")
@click.option('-f', '--force', is_flag=True, help="Force overwriting")
@click.option('-C', '--copy', is_flag=True, help="Copy password to clipboard")
@pass_db
def add(db, fullname, password, random, pattern, comment, force, copy):
    if random:
        pattern = pattern if pattern else db.config['genpass_pattern']
        password = genpass(pattern=pattern)
    elif not password:
        password = click.prompt('Password', hide_input=True, confirmation_prompt=True)

    found = db.credential(fullname=fullname)
    if found and not force:
        message = "Credential {} already exists. --force to overwrite".format(
            fullname)
        raise click.ClickException(click.style(message, fg='yellow'))

    encrypted = encrypt(password, recipient=db.config['recipient'], homedir=db.config['homedir'])
    db.add(fullname=fullname, password=encrypted, comment=comment)

    if copy:
        clipboard.copy(password)
        click.secho('Password copied to clipboard', color='yellow')

    message = 'Added {}{}'.format(fullname, ' [--force]' if force else '')
    db.repo.commit(message=message)


@cli.command(help="Copy credential password to clipboard/stdout")
@click.argument("fullname")
@click.option("--passphrase", prompt="Passphrase", hide_input=True)
@click.option("--to", default='clipboard',
              type=click.Choice(['stdout', 'clipboard']),
              help="Copy password destination")
@click.option("--clear", default=0, help="Automatically clear password from clipboard")
@pass_db
def copy(db, fullname, passphrase, to, clear):
    ensure_passphrase(passphrase, db.config)
    clear = clear if clear else db.config['copy_timeout']
    credential = db.credential(fullname)
    if not credential:
        message = "Credential '{}' not found".format(fullname)
        raise click.ClickException(click.style(message, fg='red'))

    encrypted = credential["password"]
    decrypted = decrypt(encrypted,
                        recipient=db.config['recipient'],
                        passphrase=passphrase,
                        homedir=db.config['homedir'])
    if to == 'clipboard':
        clipboard.copy(decrypted, clear)
        if not clear:
            click.secho('Password copied to clipboard', fg='yellow')
    elif to == 'stdout':
        click.echo(decrypted)


@cli.command(help="Update credential")
@click.argument("fullname")
@click.option("--name", help="Credential new name")
@click.option("--login", help="Credential new login")
@click.option("--comment", help="Credential new comment")
@click.option("--password", help="Credential new password")
@click.option('--random', is_flag=True, help="Credential new randomly generated password")
@click.option('-P', '--pattern', help="Random password regex pattern")
@pass_db
def update(db, fullname, name, login, password, random, pattern, comment):
    credential = db.credential(fullname)
    if not credential:
        message = "Credential '{}' not found".format(fullname)
        raise click.ClickException(click.style(message, fg='red'))

    if random:
        pattern = pattern if pattern else db.config['genpass_pattern']
        password = genpass(pattern=pattern)

    values = credential.copy()
    if any([name, login, password, random, comment]):
        values["name"] = name if name else credential["name"]
        values["login"] = login if login else credential["login"]
        values["password"] = password if password else credential["password"]
        values["comment"] = comment if comment else credential["comment"]
    else:
        values["name"] = click.prompt("Name", default=credential["name"])
        values["login"] = click.prompt("Login", default=credential["login"])
        values["password"] = click.prompt("Password",
                                          hide_input=True,
                                          default=credential["password"],
                                          confirmation_prompt=True,
                                          show_default=False,
                                          prompt_suffix=" [*****]: ")
        values["comment"] = click.prompt("Comment",
                                         default=credential["comment"])

    if values != credential:
        if values["password"] != credential["password"]:
            encrypted = encrypt(password, recipient=db.config['recipient'], homedir=db.config['homedir'])
            values['password'] = encrypted
        db.update(fullname=fullname, values=values)
        db.repo.commit('Updated {}'.format(credential['fullname']))


@cli.command(help="Remove credential")
@click.argument("fullname")
@click.option("-y", "--yes", is_flag=True, help="Skip confirmation prompt")
@pass_db
def remove(db, fullname, yes):
    credentials = db.credentials(fullname=fullname)

    if credentials:
        if not yes:
            creds = ', '.join([c['fullname'] for c in credentials])
            click.confirm(
                'Remove credentials: ({})'.format(
                    click.style(creds, 'yellow')),
                abort=True
            )
        for credential in credentials:
            db.remove(credential['fullname'])

        fullnames = ', '.join(c['fullname'] for c in credentials)
        db.repo.commit('Removed {}'.format(fullnames))


@cli.command(help="Search credentials by regular expressions")
@click.argument("regex")
@pass_db
def search(db, regex):
    credentials = db.matches(regex)
    if credentials:
        table = Table(
            db.config['headers'],
            table_format=db.config['table_format'],
            colors=db.config['colors'],
            hidden=['password']
        )
        click.echo(table.render(credentials))


@cli.command(help="Diagnose database for improvements")
@click.option("--full", is_flag=True, help="Show all entries")
@click.option("--days", default=90, type=int, help="Elapsed days")
@click.option("--passphrase", prompt="Passphrase", hide_input=True)
@pass_db
def status(db, full, days, passphrase):
    ensure_passphrase(passphrase, db.config)
    credentials = db.credentials()

    for cred in credentials:
        decrypted = decrypt(cred['password'],
                            recipient=db.config['recipient'],
                            passphrase=passphrase,
                            homedir=db.config['homedir'])
        cred["password"] = decrypted

    if credentials:
        limit = db.config['status_repeated_passwords_limit']
        credentials = checkers.repeated(credentials, limit)
        credentials = checkers.modified(credentials, days)

        for c in credentials:
            if c['repeated']:
                c['repeated'] = click.style(str(c['repeated']), 'red')
            if c['modified']:
                c['modified'] = click.style(str(c['modified']), 'red')

        table = Table(['fullname', 'repeated', 'modified'],
                      table_format=db.config['table_format'],
                      missing=click.style('OK', 'green'))
        click.echo(table.render(credentials))


@cli.command(name="import", help="Import credentials from path")
@click.argument("filepath", type=click.Path())
@pass_db
def import_database(db, filepath):
    importer = importers.find_importer(filepath)
    if importer:
        credentials = importer.handle(filepath)
        for cred in credentials:
            encrypted = encrypt(cred['password'], recipient=db.config['recipient'], homedir=db.config['homedir'])
            cred['password'] = encrypted
        db.insert_multiple(credentials)

        db.repo.commit(message='Imported credentials from {}'.format(filepath))


@cli.command(name="export", help="Export credentials in plain text")
@click.argument("filepath", type=click.File("w"))
@click.option("--json", "as_json", is_flag=True, help="Export as JSON")
@click.option("--passphrase", prompt="Passphrase", hide_input=True)
@pass_db
def export_database(db, filepath, as_json, passphrase):
    ensure_passphrase(passphrase, db.config)
    credentials = db.all()

    for cred in credentials:
        decrypted = decrypt(cred['password'],
                            recipient=db.config['recipient'],
                            passphrase=passphrase,
                            homedir=db.config['homedir'])
        cred["password"] = decrypted

    if as_json:
        for cred in credentials:
            cred["modified"] = str(cred["modified"])
        dict_content = {
            'handler': 'passpie',
            'version': 1.0,
            'credentials': [dict(x) for x in credentials],
        }
        content = json.dumps(dict_content, indent=2)
    else:
        dict_content = {
            'handler': 'passpie',
            'version': 1.0,
            'credentials': [dict(x) for x in credentials],
        }
        content = yaml.dump(dict_content, default_flow_style=False)

    filepath.write(content)


@cli.command(help='Renew passpie database and re-encrypt credentials')
@click.option("--passphrase", prompt="Passphrase", hide_input=True)
@pass_db
def reset(db, passphrase):
    ensure_passphrase(passphrase, db.config)
    credentials = db.credentials()
    if credentials:
        # decrypt all credentials
        for cred in credentials:
            decrypted = decrypt(cred['password'],
                                recipient=db.config['recipient'],
                                passphrase=passphrase,
                                homedir=db.config['homedir'])
            cred["password"] = decrypted

        # recreate keys if exists
        if db.has_keys():
            new_passphrase = click.prompt('New passphrase',
                                          hide_input=True,
                                          confirmation_prompt=True)
            create_keys(new_passphrase)

        # encrypt passwords
        for cred in credentials:
            cred['password'] = encrypt(cred['password'],
                                       recipient=db.config['recipient'],
                                       homedir=db.config['homedir'])

        # remove old and insert re-encrypted credentials
        db.purge()
        db.insert_multiple(credentials)

        # commit
        db.repo.commit(message='Reset database')


@cli.command(help='Remove all credentials from database')
@click.option("-y", "--yes", is_flag=True, help="Skip confirmation prompt")
@pass_db
def purge(db, yes):
    if db.credentials():
        if not yes:
            alert = "Purge '{}' credentials".format(len(db.credentials()))
            yes = click.confirm(click.style(alert, 'yellow'), abort=True)
        if yes:
            db.purge()
            db.repo.commit(message='Purged database')


@cli.command(help='Shows passpie database changes history')
@click.option("--init", is_flag=True, help="Enable history tracking")
@click.option("--reset-to", default=-1, help="Undo changes in database")
@pass_db
def log(db, reset_to, init):
    if reset_to >= 0:
        logging.info('reset database to index %s', reset_to)
        db.repo.reset(reset_to)
    elif init:
        db.repo.init()
        db.repo.commit(message='Initialized git repository', add=True)
    else:
        commits = []
        for number, message in enumerate(db.repo.commit_list()):
            number = click.style(str(number), fg='magenta')
            message = message.strip()
            commits.append("[{}] {}".format(number, message))

        for commit in reversed(commits):
            print(commit)



# -*- coding: utf-8 -*-
import re

import pytest

import passpie.crypt
from passpie.crypt import (
    KEY_INPUT,
    DEVNULL,
    make_key_input,
    export_keys,
    import_keys,
    create_keys,
)


@pytest.fixture
def mock_call(mocker):
    return mocker.patch('passpie.crypt.process.call')

@pytest.fixture
def mock_open():
    try:
        from mock import mock_open as mopen
    except:
        from unittest.mock import mock_open as mopen
    return mopen()


def test_crypt_make_key_input_handles_unicode_encode_error_handling(mocker):
    passphrase = 'passphrase'
    key_length = '2064'
    key_input = KEY_INPUT.format(key_length, passphrase)
    assert key_input == make_key_input(passphrase, key_length)


def test_crypt_make_key_input_handles_unicode_encode_error(mocker):
    passphrase = u"L'éphémère"
    key_length = '2064'
    key_input = make_key_input(passphrase, key_length)

    assert key_input is not None


def test_crypt_export_keys_calls_gpg_command_on_export_keys(mocker):
    output = 'command output'
    mock_call = mocker.patch('passpie.crypt.process.call', return_value=(output, ''))
    mocker.patch('passpie.crypt.which', return_value='gpg')
    homedir = 'mock_homedir'

    secret = False
    command = [
        'gpg',
        '--no-version',
        '--batch',
        '--homedir', homedir,
        '--export-secret-keys' if secret else '--export',
        '--armor',
        '-o', '-'
    ]
    result = export_keys(homedir)

    assert mock_call.called is True
    assert result == output
    mock_call.assert_called_once_with(command)


def test_create_keys_export_public_and_secret_keys_into_stdout(mocker):
    passphrase = "passphrase"
    homedir = "mock/homedir"
    key_input = 'some key input'
    output = '--GPG--'
    mocker.patch('passpie.crypt.which', return_value='gpg')
    mocker.patch('passpie.crypt.make_key_input', return_value=key_input)
    mock_call = mocker.patch('passpie.crypt.process.call',
                             return_value=(output, ''))
    mock_tempdir = mocker.patch('passpie.crypt.tempdir', return_value=homedir)

    command = [
        'gpg',
        '--batch',
        '--no-tty',
        '--no-secmem-warning',
        '--no-permission-warning',
        '--no-mdc-warning',
        '--homedir', homedir,
        '--gen-key',
    ]

    result = create_keys(passphrase)

    assert result == output
    assert mock_call.called
    mock_call.assert_called_once_with(command, input=key_input)


def test_create_keys_if_path_is_passed_create_file_homedir(mocker, mock_call, mock_open):
    mock_call.return_value = ('', '')
    mock_tempdir = mocker.patch('passpie.crypt.tempdir', mock_open(), create=True)
    mocker.patch('passpie.crypt.export_keys', side_effect=['PUBLIC', 'PRIVATE'])
    mock_open = mocker.patch("passpie.crypt.open", mock_open(), create=True)
    mock_keysfile = mock_open().__enter__()
    mock_tempdir().__enter__().return_value = 'create_keys'

    create_keys('passphrase', 'path')
    assert mock_keysfile.write.called
    mock_keysfile.write.assert_any_call('PUBLIC')
    mock_keysfile.write.assert_any_call('PRIVATE')


def test_encrypt_calls_gpg_encrypt_command_with_recipient(mocker, mock_call):
    recipient = 'passpie@local'
    password = 's3cr3t'
    mock_call.return_value = ('--GPG ENCRYPTED--', None)
    homedir = 'homedir'
    mocker.patch('passpie.crypt.which', return_value='gpg')
    command = [
        'gpg',
        '--batch',
        '--no-tty',
        '--always-trust',
        '--armor',
        '--recipient', recipient,
        '--homedir', homedir,
        '--encrypt'
    ]
    result = passpie.crypt.encrypt(password, recipient, homedir)

    assert result is not None
    assert mock_call.called
    mock_call.assert_called_once_with(command, input=password)


def test_decrypt_calls_gpg_encrypt_expected_command(mocker, mock_call):
    recipient = 'passpie@local'
    passphrase = 'passphrase'
    homedir = 'homedir'
    data = '--GPG ENCRYPTED--'
    mock_call.return_value = ('s3cr3t', None)
    mocker.patch('passpie.crypt.which', return_value='gpg')
    command = [
        'gpg',
        '--batch',
        '--no-tty',
        '--always-trust',
        '--recipient', recipient,
        '--homedir', homedir,
        '--passphrase', passphrase,
        '--emit-version',
        '-o', '-',
        '-d', '-',
    ]
    result = passpie.crypt.decrypt(data, recipient, passphrase, homedir)

    assert result is not None
    assert mock_call.called
    mock_call.assert_called_once_with(command, input=data)


def test_default_recipient_returns_first_matched_fingerprint(mocker, mock_call):
    output = '123\n456'
    mocker.patch('passpie.crypt.tempdir')
    mocker.patch('passpie.crypt.import_keys')
    mocker.patch('passpie.crypt.re.search',
                 return_value=re.match('\d+', output))
    mock_call.return_value = (output, '')

    recipient = passpie.crypt.get_default_recipient('homedir', secret=False)
    assert recipient == '123'


def test_default_recipient_calls_expected_command_when_secret(mocker, mock_call):
    mocker.patch('passpie.crypt.tempdir')
    mocker.patch('passpie.crypt.which', return_value='gpg')
    mocker.patch('passpie.crypt.import_keys')
    output = "123\n456"
    mock_call.return_value = (output, '')
    recipient = passpie.crypt.get_default_recipient('homedir', secret=True)

    command = [
        'gpg',
        '--no-tty',
        '--batch',
        '--no-secmem-warning',
        '--no-permission-warning',
        '--no-mdc-warning',
        '--list-{}-keys'.format('secret'),
        '--fingerprint',
        '--homedir', 'homedir',
    ]

    assert mock_call.called is True
    mock_call.assert_called_once_with(command)



from passpie import checkers
from datetime import datetime, timedelta


def test_repeated_deepcopy_credentials(mocker):
    mock_deepcopy = mocker.patch('passpie.checkers.deepcopy')
    credentials = [{'fullname': 'foo@example.com', 'password': 's3cr3t'}]

    checkers.repeated(credentials, limit=1)
    assert mock_deepcopy.called
    mock_deepcopy.assert_called_once_with(credentials)


def test_modified_sets_credential_modifield_field_none(mocker):
    credentials = [{'modified': datetime(year=2016, month=1, day=1)}]
    expected_credentials = [{'modified': None}]
    mock_datetime = mocker.patch('passpie.checkers.datetime')
    mock_datetime.now.return_value = credentials[0]['modified'] + timedelta(days=1)

    result_credentials = checkers.modified(credentials, days=1)
    assert result_credentials == expected_credentials


def test_modified_sets_credential_modifield_field__with_string_days_count(mocker):
    credentials = [{'modified': datetime(year=2016, month=1, day=1)}]
    expected_credentials = [{'modified': "3 days ago"}]
    mock_datetime = mocker.patch('passpie.checkers.datetime')
    mock_datetime.now.return_value = credentials[0]['modified'] + timedelta(days=3)

    result_credentials = checkers.modified(credentials, days=1)
    assert result_credentials == expected_credentials


def test_modified_deepcopy_credentials(mocker):
    mock_deepcopy = mocker.patch('passpie.checkers.deepcopy')
    credentials = [{'modified': datetime.now()}]

    checkers.modified(credentials, days=1)
    assert mock_deepcopy.called
    mock_deepcopy.assert_called_once_with(credentials)



from datetime import datetime
import logging
import os
import re
import shutil

from tinydb import TinyDB, Storage, where, Query
import yaml

from .utils import mkdir_open
from .history import Repository
from .credential import split_fullname, make_fullname


def is_repo_url(path):
    if path:
        return re.match(
            r'((git|ssh|http(s)?)|(git@[\w\.]+))(:(//)?)([\w\.@\:/\-~]+)(\.git)(/)?',
            path
        )


class PasspieStorage(Storage):
    extension = ".pass"

    def __init__(self, path):
        super(PasspieStorage, self).__init__()
        self.path = path

    def delete(self, credentials):
        for cred in credentials:
            dirname, filename = cred["name"], cred["login"] + self.extension
            credpath = os.path.join(self.path, dirname, filename)
            os.remove(credpath)
            if not os.listdir(os.path.dirname(credpath)):
                shutil.rmtree(os.path.dirname(credpath))

    def read(self):
        elements = []
        for rootdir, dirs, files in os.walk(self.path):
            filenames = [f for f in files if f.endswith(self.extension)]
            for filename in filenames:
                docpath = os.path.join(rootdir, filename)
                with open(docpath) as f:
                    elements.append(yaml.load(f.read()))

        return {"_default":
                {idx: elem for idx, elem in enumerate(elements, start=1)}}

    def write(self, data):
        deleted = [c for c in self.read()["_default"].values()
                   if c not in data["_default"].values()]
        self.delete(deleted)

        for eid, cred in data["_default"].items():
            dirname, filename = cred["name"], cred["login"] + self.extension
            credpath = os.path.join(self.path, dirname, filename)
            with mkdir_open(credpath, "w") as f:
                f.write(yaml.dump(dict(cred), default_flow_style=False))


class Database(TinyDB):

    def __init__(self, config, storage=PasspieStorage):
        self.config = config
        self.path = config['path']
        self.repo = Repository(self.path,
                               autopull=config.get('autopull'),
                               autopush=config.get('autopush'))
        PasspieStorage.extension = config['extension']
        super(Database, self).__init__(self.path, storage=storage)

    def has_keys(self):
        return os.path.exists(os.path.join(self.path, '.keys'))

    def credential(self, fullname):
        login, name = split_fullname(fullname)
        return self.get((where("login") == login) & (where("name") == name))

    def add(self, fullname, password, comment):
        login, name = split_fullname(fullname)
        if login is None:
            logging.error('Cannot add credential with empty login. use "@<name>" syntax')
            return None
        credential = dict(fullname=fullname,
                          name=name,
                          login=login,
                          password=password,
                          comment=comment,
                          modified=datetime.now())
        self.insert(credential)
        return credential

    def update(self, fullname, values):
        values['fullname'] = make_fullname(values["login"], values["name"])
        values['modified'] = datetime.now()
        self.table().update(values, (where("fullname") == fullname))

    def credentials(self, fullname=None):
        if fullname:
            login, name = split_fullname(fullname)
            Credential = Query()
            if login is None:
                creds = self.search(Credential.name == name)
            else:
                creds = self.search((Credential.login == login) & (Credential.name == name))
        else:
            creds = self.all()
        return sorted(creds, key=lambda x: x["name"] + x["login"])

    def remove(self, fullname):
        self.table().remove(where('fullname') == fullname)

    def matches(self, regex):
        Credential = Query()
        credentials = self.search(
            Credential.name.matches(regex) |
            Credential.login.matches(regex) |
            Credential.comment.matches(regex)
        )
        return sorted(credentials, key=lambda x: x["name"] + x["login"])



"""
parts of this code from pyperclip: https://github.com/asweigart/pyperclip
"""
import ctypes
import logging
import platform
import time

from . import process
from ._compat import *


text_type = unicode if is_python2() else str

LINUX_COMMANDS = {
    'xsel': ['xsel', '-ibps'],
    'xclip': ['xclip', '-i']
}

OSX_COMMANDS = {
    'pbcopy': ['pbcopy', 'w']
}


def ensure_commands(commands):
    for command_name, command in commands.items():
        if which(command_name) and command:
            return command
    else:
        logging.error('missing commands: ', ' or '.join(commands))


def clean(command, delay):
    for dot in ['.' for _ in range(delay)]:
        sys.stdout.write(dot)
        sys.stdout.flush()
        time.sleep(1)
    else:
        process.call(command, input='\b')
        print('')


def _copy_windows(text, clear=0):
    GMEM_DDESHARE = 0x2000
    CF_UNICODETEXT = 13
    d = ctypes.windll  # cdll expects 4 more bytes in user32.OpenClipboard(0)
    if not isinstance(text, text_type):
        text = text.decode('mbcs')

    d.user32.OpenClipboard(0 if is_python2() else None)

    d.user32.EmptyClipboard()
    hCd = d.kernel32.GlobalAlloc(GMEM_DDESHARE, len(text.encode('utf-16-le')) + 2)
    pchData = d.kernel32.GlobalLock(hCd)
    ctypes.cdll.msvcrt.wcscpy(ctypes.c_wchar_p(pchData), text)
    d.kernel32.GlobalUnlock(hCd)
    d.user32.SetClipboardData(CF_UNICODETEXT, hCd)
    d.user32.CloseClipboard()


def _copy_cygwin(text, clear=0):
    GMEM_DDESHARE = 0x2000
    CF_UNICODETEXT = 13
    d = ctypes.cdll
    if not isinstance(text, text_type):
        text = text.decode('mbcs')
    d.user32.OpenClipboard(0)
    d.user32.EmptyClipboard()
    hCd = d.kernel32.GlobalAlloc(GMEM_DDESHARE,
                                 len(text.encode('utf-16-le')) + 2)
    pchData = d.kernel32.GlobalLock(hCd)
    ctypes.cdll.msvcrt.wcscpy(ctypes.c_wchar_p(pchData), text)
    d.kernel32.GlobalUnlock(hCd)
    d.user32.SetClipboardData(CF_UNICODETEXT, hCd)
    d.user32.CloseClipboard()


def _copy_osx(text, clear=0):
    command = ensure_commands(OSX_COMMANDS)
    process.call(command, input=text)
    if clear:
        clean(command, delay=clear)


def _copy_linux(text, clear=0):
    command = ensure_commands(LINUX_COMMANDS)
    process.call(command, input=text)
    if clear:
        clean(command, delay=clear)


def copy(text, clear=0):
    platform_name = platform.system().lower()
    if platform_name == 'darwin':
        _copy_osx(text, clear)
    elif platform_name == 'linux':
        _copy_linux(text, clear)
    elif platform_name == 'windows':
        _copy_windows(text, clear)
    elif 'cygwin' in platform_name.lower():
        _copy_cygwin(text, clear)
    else:
        msg = "platform '{}' copy to clipboard not supported".format(
            platform_name)
        logging.error(msg)
        return
    logging.debug('text copied to clipboard')



from copy import deepcopy
from datetime import datetime, timedelta


def repeated(credentials, limit):
    result_credentials = deepcopy(credentials)
    for i, cred in enumerate(credentials):
        rep = [c['fullname'] for c in credentials
               if cred['password'] == c['password'] and
               cred['fullname'] != c['fullname']][:limit]
        if not rep:
            rep = None
        result_credentials[i]['repeated'] = rep
    return result_credentials


def modified(credentials, days):
    result_credentials = deepcopy(credentials)
    for i, cred in enumerate(credentials):
        modified_delta = (datetime.now() - cred["modified"])
        if modified_delta > timedelta(days=days):
            modified_time = "{} days ago".format(modified_delta.days)
        else:
            modified_time = None
        result_credentials[i]['modified'] = modified_time
    return result_credentials



import click
from click.testing import CliRunner
import pytest

from passpie.cli import cli, ensure_passphrase


@pytest.fixture
def mock_deps(mocker):
    return mocker.patch('passpie.cli.ensure_dependencies')


def test_ensure_passphrase_calls_decrypt_with_encrypted_data(mocker, mock_deps):
    mock_logging = mocker.patch('passpie.cli.logging')
    mock_encrypt = mocker.patch('passpie.cli.encrypt')
    mock_decrypt = mocker.patch('passpie.cli.decrypt', return_value='OK')
    config = {'recipient': 'recipient', 'homedir': 'homedir'}

    ensure_passphrase('passphrase', config=config)
    assert mock_encrypt.called
    assert mock_decrypt.called
    mock_encrypt.assert_called_once_with('OK',
                                         recipient=config['recipient'],
                                         homedir=config['homedir'])
    mock_decrypt.assert_called_once_with(mock_encrypt.return_value,
                                         recipient=config['recipient'],
                                         passphrase='passphrase',
                                         homedir=config['homedir'])


def test_ensure_passphrase_logs_error_when_decrypt_result_not_ok(mocker, mock_deps):
    mock_logging = mocker.patch('passpie.cli.logging')
    mock_encrypt = mocker.patch('passpie.cli.encrypt')
    mock_decrypt = mocker.patch('passpie.cli.decrypt', return_value='NOT OK')
    config = {'recipient': 'recipient', 'homedir': 'homedir'}
    message_full = "Wrong passphrase for recipient: {} in homedir: {}".format(
        config['recipient'],
        config['homedir'],
    )

    with pytest.raises(click.ClickException):
        ensure_passphrase('passphrase', config=config)

    assert mock_logging.error.called
    mock_logging.error.assert_called_once_with(message_full)


def test_call_to_cli_exit_with_error_when_missing_dependencies(mocker):
    mocker.patch('passpie.cli.Database')
    mocker.patch('passpie.cli.config')
    mocker.patch('passpie.cli.ensure_dependencies', side_effect=RuntimeError)
    mocker.patch('passpie.cli.logging')

    runner = CliRunner()
    result = runner.invoke(cli)
    assert result.exit_code != 0


def test_cli_create_database_with_configuration(mocker, mock_deps):
    mock_database = mocker.patch('passpie.cli.Database')
    mock_config = mocker.patch('passpie.cli.config')
    mocker.patch('passpie.cli.logging')

    runner = CliRunner()
    result = runner.invoke(cli)

    assert mock_database.called
    mock_database.assert_called_once_with(mock_config.load())


def test_cli_create_database_with_configuration_overriding_autopush(mocker, mock_deps):
    mock_database = mocker.patch('passpie.cli.Database')
    mock_config = mocker.patch('passpie.cli.config')
    mocker.patch('passpie.cli.logging')

    runner = CliRunner()
    result = runner.invoke(cli, ['--autopush', 'origin/master'])

    assert mock_config.load.called
    mock_config.load.assert_called_once_with(autopush=('origin', 'master'))


def test_cli_create_database_with_configuration_overriding_autopull(mocker, mock_deps):
    mock_database = mocker.patch('passpie.cli.Database')
    mock_config = mocker.patch('passpie.cli.config')
    mocker.patch('passpie.cli.logging')

    runner = CliRunner()
    result = runner.invoke(cli, ['--autopull', 'origin/master'])

    assert mock_config.load.called
    mock_config.load.assert_called_once_with(autopull=('origin', 'master'))


def test_cli_sets_logging_verbose_level_to_info_when_passing_one_v(mocker, mock_deps):
    mocker.patch('passpie.cli.Database')
    mocker.patch('passpie.cli.config')
    mock_logging = mocker.patch('passpie.cli.logging')

    runner = CliRunner()
    result = runner.invoke(cli, ['-v'])

    assert mock_logging.basicConfig.called
    _, kwargs = mock_logging.basicConfig.call_args
    assert kwargs['level'] == mock_logging.INFO
    assert result.exit_code == 0


def test_cli_sets_logging_verbose_level_to_debug_when_passing_two_v(mocker, mock_deps):
    mocker.patch('passpie.cli.Database')
    mocker.patch('passpie.cli.config')
    mock_logging = mocker.patch('passpie.cli.logging')

    runner = CliRunner()
    result = runner.invoke(cli, ['-vv'])

    assert result.exit_code == 0
    assert mock_logging.basicConfig.called
    _, kwargs = mock_logging.basicConfig.call_args
    assert kwargs['level'] == mock_logging.DEBUG


def test_cli_sets_logging_verbose_level_to_critical_when_no_verbose_passed(mocker, mock_deps):
    mocker.patch('passpie.cli.Database')
    mocker.patch('passpie.cli.config')
    mock_logging = mocker.patch('passpie.cli.logging')

    runner = CliRunner()
    result = runner.invoke(cli)

    assert result.exit_code == 0
    assert mock_logging.basicConfig.called
    _, kwargs = mock_logging.basicConfig.call_args
    assert kwargs['level'] == mock_logging.CRITICAL



import os
import sys
try:
    from shutil import which as _which
except ImportError:
    from distutils.spawn import find_executable as _which

try:
    basestring
except NameError:
    basestring = str

try:
    unicode = unicode
except NameError:
    unicode = str


def which(binary):
    path = _which(binary)
    if path:
        realpath = os.path.realpath(path)
        return realpath
    return None


def is_python2():
    return sys.version_info < (3,)


class FileNotFoundError(OSError):

    def __init__(self, message="No such file or directory"):
        super(FileNotFoundError, self).__init__(2, message)


class FileExistsError(OSError):

    def __init__(self, message="File exists"):
        super(FileExistsError, self).__init__(17, message)



from contextlib import contextmanager
import errno
import os
import tempfile

import rstr

from ._compat import which


import_module = __import__


def genpass(pattern=r'[\w]{32}'):
    """generates a password with random chararcters
    """
    return rstr.xeger(pattern)


@contextmanager
def mkdir_open(path, mode="r"):
    try:
        dir_path = os.path.dirname(path)
        os.makedirs(dir_path)
    except OSError as exc:
        if exc.errno == errno.EEXIST and os.path.isdir(dir_path):
            pass
        else:
            raise
    with open(path, mode) as fd:
        yield fd


def ensure_dependencies():
    try:
        assert which('gpg') or which('gpg2')
    except AssertionError:
        raise RuntimeError('GnuPG not installed. https://www.gnupg.org/')


def tempdir():
    return tempfile.mkdtemp()


def touch(path):
    with open(path, "w"):
        pass



#!/usr/bin/env python

import os
import sys

try:
    from setuptools import setup, Command, find_packages
except ImportError:
    from distutils.core import setup, Command, find_packages


__version__ = "1.1.0"
long_description = open('README.md').read() + "\n\n"


if sys.argv[-1] == 'register':
    import pypandoc
    long_description = pypandoc.convert('README.md', 'rst')


if sys.argv[-1] == 'publish':
    os.system('python setup.py sdist upload')
    os.system('python setup.py bdist_wheel upload')
    sys.exit()


if sys.argv[-1] == 'tag':
    os.system("git tag -a v%s -m 'version v%s'" % (__version__, __version__))
    os.system("git push --tags")
    sys.exit()


requirements = [
    'click==6.2',
    'PyYAML==3.11',
    'tabulate==0.7.5',
    'tinydb==3.1.1',
    'rstr==2.2.3',
]


class PyTest(Command):
    user_options = [('pytest-args=', 'a', "Arguments to pass to py.test")]

    def initialize_options(self):
        self.pytest_args = []

    def finalize_options(self):
        pass

    def run(self):
        import pytest
        errno = pytest.main(self.pytest_args)
        sys.exit(errno)


class PyTestCoverage(PyTest):

    def initialize_options(self):
        self.pytest_args = ['--cov', 'passpie']


setup(
    name='passpie',
    version=__version__,
    license='License :: OSI Approved :: MIT License',
    description="Manage your login credentials from the terminal painlessly.",
    long_description=long_description,
    author='Marcwebbie',
    author_email='marcwebbie@gmail.com',
    url='https://github.com/marcwebbie/passpie',
    download_url='https://github.com/marcwebbie/passpie',
    packages=find_packages(),
    entry_points={
        'console_scripts': [
            'passpie=passpie.cli:cli',
        ]
    },
    install_requires=requirements,
    cmdclass={'test': PyTest, 'coverage': PyTestCoverage},
    test_suite='tests',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: End Users/Desktop',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Programming Language :: Python',
        'Topic :: Security :: Cryptography',
    ],
)



import logging
import os
from subprocess import Popen, PIPE

from ._compat import *


DEVNULL = open(os.devnull, 'w')


class Proc(Popen):

    def communicate(self, **kwargs):
        if kwargs.get('input') and isinstance(kwargs['input'], basestring):
            kwargs['input'] = kwargs['input'].encode('utf-8')
        return super(Proc, self).communicate(**kwargs)

    def __exit__(self, *args, **kwargs):
        if hasattr(super(Proc, self), '__exit__'):
            super(Proc, self).__exit__(*args, **kwargs)

    def __enter__(self, *args, **kwargs):
        if hasattr(super(Proc, self), '__enter__'):
            print(super(Proc, self).__enter__(*args, **kwargs))
            return super(Proc, self).__enter__(*args, **kwargs)
        return self


def call(*args, **kwargs):
    if logging.getLogger().getEffectiveLevel() == logging.DEBUG:
        stderr = PIPE
    else:
        stderr = DEVNULL

    kwargs.setdefault('stderr', stderr)
    kwargs.setdefault('stdout', PIPE)
    kwargs.setdefault('stdin', PIPE)
    kwargs.setdefault('shell', False)
    kwargs_input = kwargs.pop('input', None)

    with Proc(*args, **kwargs) as proc:
        output, error = proc.communicate(input=kwargs_input)
        try:
            output = output.decode('utf-8')
            error = error.decode('utf-8')
        except AttributeError:
            pass
        return output, error



import os
import re

from . import process
from .utils import tempdir
from ._compat import *

from passpie.utils import which


GPG_HOMEDIR = os.path.expanduser('~/.gnupg')
DEVNULL = open(os.devnull, 'w')
KEY_INPUT = u"""Key-Type: RSA
Key-Length: {}
Subkey-Type: RSA
Name-Comment: Auto-generated by Passpie
Passphrase: {}
Name-Real: Passpie
Name-Email: passpie@local
Expire-Date: 0
%commit
"""


def ensure_keys(path):
    keys_path = os.path.join(os.path.expanduser(path), '.keys')
    if os.path.isfile(keys_path):
        return keys_path


def make_key_input(passphrase, key_length):
    passphrase = unicode(passphrase)
    key_length = unicode(key_length)
    key_input = KEY_INPUT.format(key_length, passphrase)
    return key_input


def export_keys(homedir, secret=False):
    command = [
        which('gpg2') or which('gpg'),
        '--no-version',
        '--batch',
        '--homedir', homedir,
        '--export-secret-keys' if secret else '--export',
        '--armor',
        '-o', '-'
    ]
    output, error = process.call(command)
    return output


def create_keys(passphrase, path=None, key_length=4096):
    homedir = tempdir()
    command = [
        which('gpg2') or which('gpg'),
        '--batch',
        '--no-tty',
        '--no-secmem-warning',
        '--no-permission-warning',
        '--no-mdc-warning',
        '--homedir', homedir,
        '--gen-key',
    ]
    key_input = make_key_input(passphrase, key_length)
    output, error = process.call(command, input=key_input)
    if path:
        with open(path, 'w') as keysfile:
            keysfile.write(export_keys(homedir))
            keysfile.write(export_keys(homedir, secret=True))
    else:
        return output


def import_keys(keys_path, homedir):
    command = [
        which('gpg2') or which('gpg'),
        '--no-tty',
        '--batch',
        '--no-secmem-warning',
        '--no-permission-warning',
        '--no-mdc-warning',
        '--homedir', homedir,
        '--import', keys_path
    ]
    output, _ = process.call(command)
    return homedir


def get_default_recipient(homedir, secret=False):
    command = [
        which('gpg2') or which('gpg'),
        '--no-tty',
        '--batch',
        '--no-secmem-warning',
        '--no-permission-warning',
        '--no-mdc-warning',
        '--list-{}-keys'.format('secret' if secret else 'public'),
        '--fingerprint',
        '--homedir', homedir,
    ]
    output, _ = process.call(command)
    for line in output.splitlines():
        try:
            mobj = re.search(r'(([0-9A-F]{4}\s*?){10})', line)
            fingerprint = mobj.group().replace(' ', '')
            return fingerprint
        except (AttributeError, IndexError):
            continue
    return ''


def encrypt(data, recipient, homedir):
    recipient = recipient if recipient else get_default_recipient(homedir)
    command = [
        which('gpg2') or which('gpg'),
        '--batch',
        '--no-tty',
        '--always-trust',
        '--armor',
        '--recipient', recipient,
        '--homedir', homedir,
        '--encrypt'
    ]
    output, _ = process.call(command, input=data)
    return output


def decrypt(data, recipient, passphrase, homedir):
    recipient = recipient if recipient else get_default_recipient(homedir)
    command = [
        which('gpg2') or which('gpg'),
        '--batch',
        '--no-tty',
        '--always-trust',
        '--recipient', recipient,
        '--homedir', homedir,
        '--passphrase', passphrase,
        '--emit-version',
        '-o', '-',
        '-d', '-',
    ]
    output, _ = process.call(command, input=data)
    return output



from functools import wraps
import logging

from . import process
from .utils import which, tempdir


def ensure_git(return_value=None):
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            if which('git'):
                try:
                    return func(*args, **kwargs)
                except Exception as e:
                    logging.debug(str(e))
            else:
                logging.debug('git is not installed')
            return return_value
        return wrapper
    return decorator


@ensure_git()
def clone(url, dest=None):
    dest = dest if dest else tempdir()
    cmd = ['git', 'clone', url, dest]
    process.call(cmd)
    return dest


class Repository(object):

    def __init__(self, path, autopull=None, autopush=None):
        self.path = path
        self.autopush = autopush
        self.autopull = autopull
        if autopull:
            self.pull_rebase(*autopull)

    @ensure_git()
    def init(self):
        cmd = ['git', 'init', self.path]
        process.call(cmd, cwd=self.path)

    @ensure_git()
    def pull_rebase(self, remote='origin', branch='master'):
        cmd = ['git', 'pull', '--rebase', remote, branch]
        process.call(cmd, cwd=self.path)

    @ensure_git()
    def push(self, remote='origin', branch='master'):
        cmd = ['git', 'push', remote, branch]
        process.call(cmd, cwd=self.path)

    @ensure_git()
    def add(self, all=False):
        if all is True:
            cmd = ['git', 'add', '--all', '.']
        else:
            cmd = ['git', 'add', '.']
        process.call(cmd, cwd=self.path)

    @ensure_git()
    def commit(self, message, add=True):
        if add:
            self.add(all=True)
        cmd = ['git', 'commit', '-m', message]
        process.call(cmd, cwd=self.path)
        if self.autopush:
            self.push()

    @ensure_git(return_value=[])
    def commit_list(self):
        cmd = ['git', 'log', '--reverse', '--pretty=format:%s']
        output, _ = process.call(cmd, cwd=self.path)
        return output.splitlines()

    @ensure_git(return_value=[])
    def sha_list(self):
        cmd = ['git', 'log', '--reverse', '--pretty=format:%h']
        output, _ = process.call(cmd, cwd=self.path)
        return output.splitlines()

    @ensure_git()
    def reset(self, to_index):
        try:
            sha = self.sha_list()[to_index]
        except IndexError:
            logging.info('commit on index "{}" not found'.format(to_index))

        cmd = ['git', 'reset', '--hard', sha]
        process.call(cmd, cwd=self.path)



import copy
import logging
import os

import yaml

from .utils import tempdir
from .crypt import ensure_keys, import_keys, get_default_recipient


DEFAULT_PATH = os.path.join(os.path.expanduser('~/.passpierc'))
DEFAULT = {
    'path': os.path.join(os.path.expanduser('~/.passpie')),
    'short_commands': False,
    'key_length': 4096,
    'genpass_pattern': r'[a-z]{10} [-_+=*&%$#]{10} [A-Z]{10}',
    'homedir': os.path.join(os.path.expanduser('~/.gnupg')),
    'recipient': None,
    'table_format': 'fancy_grid',
    'headers': ['name', 'login', 'password', 'comment'],
    'colors': {'name': 'yellow', 'login': 'green'},
    'repo': True,
    'autopull': None,
    'autopush': None,
    'status_repeated_passwords_limit': 5,
    'copy_timeout': 0,
    'extension': '.pass',
    'recipient': None
}


def read(path):
    try:
        with open(path) as config_file:
            content = config_file.read()
        config = yaml.load(content)
    except IOError:
        logging.debug('config file "%s" not found' % path)
        return DEFAULT
    except yaml.scanner.ScannerError:
        logging.error('Malformed user configuration file: {}'.format(path))
        return {}

    return config


def read_global_config():
    return read(DEFAULT_PATH)


def create(path, defaults={}, filename='.config'):
    config_path = os.path.join(os.path.expanduser(path), filename)
    with open(config_path, 'w') as config_file:
        config_file.write(yaml.dump(defaults, default_flow_style=False))


def setup_crypt(configuration):
    keys_filepath = ensure_keys(configuration['path'])
    if keys_filepath:
        configuration['homedir'] = tempdir()
        import_keys(keys_filepath, configuration['homedir'])
    if not configuration['recipient']:
        configuration['recipient'] = get_default_recipient(configuration['homedir'])
    return configuration


def load(**overrides):
    database_path = overrides.get('path', DEFAULT['path'])
    local_config_path = os.path.join(os.path.expanduser(database_path), '.config')
    configuration = copy.deepcopy(DEFAULT)
    if os.path.exists(DEFAULT_PATH):
        configuration.update(read_global_config())
    if os.path.exists(local_config_path):
        configuration.update(read(local_config_path))
    if overrides:
        configuration.update(overrides)

    configuration = setup_crypt(configuration)
    return configuration



from tabulate import tabulate
import click


class Table(object):

    def __init__(self, headers, table_format='rst', colors=None, hidden=None, missing=None):
        self.headers = headers
        self.colors = colors if colors else {}
        self.hidden = hidden if hidden else []
        self.table_format = table_format
        self.missing = missing

    def colorize(self, key, text):
        return click.style(text, fg=self.colors.get(key))

    def render(self, data):
        data = sorted(data, key=lambda c: c[self.headers[0]])
        rows = []
        for entry in data:
            row = []
            for header in self.headers:
                if header in self.hidden:
                    entry[header] = '*****'
                elif header in self.colors:
                    text = entry[header]
                    entry[header] = self.colorize(header, text)
                row.append(entry[header])
            rows.append(row)

        headers = [click.style(h.title(), bold=True) for h in self.headers]
        if rows:
            return tabulate(rows,
                            headers,
                            tablefmt=self.table_format,
                            missingval=self.missing,
                            numalign='left')
